# Copyright Didier Bur - 2004
# A personal license is granted to you to use this program freely
# You are not authorized to modify or copy this script
# to use at ends of trade
# This license does not imply any guarantee
# You use this program at your own risks
# For more information, please contact : bur@crai.archi.fr

#--------------------------------------------------------------------------
# Import a points clouds in SU, assuming the format of the input file is:
# x delimiter y delimiter z
# for instance: 1.0,2.0,3.0 (a CSV file format in this case)
#--------------------------------------------------------------------------
# Versions history:
# 30.07.04: first release, just reads the points
# 01.08.04: some delimiters added, layer selectable
# 18.06.07: better and single dialog box (more user-friendly).
#           better layer support
#           more delimiters, sample of file line for reference
#           invalid data file detection, more than 3 data per line detection
#           select XYZ locations option
#           flatten Z ccordinate option
#           zoom extend at command end
#           ScriptOrganizer compatible
#           Triangulation option
#--------------------------------------------------------------------------

# Supports Organizer.rb
#$submenu ? (organizer = $submenu) : (organizer = UI.menu("Plugins"))
#organizer.add_item("Import points cloud") { import_points_cloud }

require 'sketchup.rb'

def import_points_cloud
	
  model = Sketchup.active_model
  saved_layer = model.active_layer
  
  opt = Sketchup.active_model.options
  ###u = opt[0][2]
  u=opt["UnitsOptions"]["LengthUnit"]###TIG fix

case u
  when 0
    $mult = 1
  when 1
    $mult = 12
  when 2
    $mult = 0.0393700787401
  when 3
    $mult = 0.393700787401
  when 4
    $mult = 39.3700787401
end    

comp_list = collect_components
comp_list.push "Construction point"
lay_list = collect_layers
start_layer = model.active_layer

value = UI.openpanel "Select points cloud file" 
if value
  points_cloud_File = File.new( value.to_s, "r" )                             #opens the input file
  tableau_lignes = IO.readlines(value.to_s)
  tableau_lignes.uniq!
  npts = tableau_lignes.length
  Sketchup.set_status_text(npts.to_s + " points to import.\n")
    
#----------------------------------------------------------------------------- Dialog box
    # Delimiter settings
    @delim_options = "," if not @delim_options
    $point_type = "Construction point" if not $point_type
    flatten = "No" if not flatten
    delim_options = ["SPACE",",",";",".","#","-","_","!","?","/",":","'","*"]
    enum1 = [delim_options.join("|")]
    enum2 = [comp_list.join("|")]
    enum3 = [lay_list.join("|")]
    enum4 = [["Yes", "No"].join("|")]
    enums=[enum1,enum2,enum4,enum3]
    new_layer = ""
    prompts = ["Data delimiter: ", "Point object: ", "Flatten Z coordinate:", "Put points on existing layer: ", "Put points on new layer: ", "Line sample for reference: "]
    values = [@delim_options, $point_type, flatten, lay_list[0], "", tableau_lignes[0]]
    
    results = inputbox prompts, values, enums, "Points Cloud Options (" + npts.to_s + " points to import)."
    return nil if not results
    @delim_options = results[0]
    $point_type = results[1]
    layer = results[3]
    flatten = results[2]
    new_layer = results[4]
    
    # Delimiter exceptions
    if( @delim_options == "SPACE" ) 
      @delim_options = " " 
    end
    
    sample_line = tableau_lignes[0].split(@delim_options)
    # Not enough data per line
    if sample_line.length < 3
      UI.messagebox("Only " + sample_line.length.to_s + " data per line. Aborting...")
      return nil
    end
    # More than XYZ coords per line:
    if sample_line.length > 3
      UI.messagebox("More than 3 XYZ coordinates per line (" + sample_line.length.to_s + ").\n Please select X, Y, and Z in the following line sample.")
      elements_list=["Indice","X","Y","Z","Nothing", "Nothing"]
      enum = [elements_list.join("|")]
      enums = [enum, enum, enum, enum, enum, enum]
      prompts = ["First element: ", "Second element: ", "Third element: ", "Fourth element", "Fifth element","Sixth element: ", "Line sample: "]
      values=[elements_list[0], elements_list[1],elements_list[2],elements_list[3],elements_list[4], elements_list[5], tableau_lignes[0]]
      @order = inputbox prompts, values, enums, "Select coordinates location in line"
      indx= @order.index("X")
      indy= @order.index("Y")
      indz= @order.index("Z")
      return nil if not @order

    end
    
    # Delimiter exceptions
    if( @delim_options == "SPACE" ) 
      @delim_options = " " 
    end
    #if( @delim_options == "TAB" ) 
      #@delim_options = "\t" 
    #end
    
    # Layer
    if new_layer == ""
      model.active_layer = layer
    else
      model.layers.add(new_layer)
      model.active_layer = new_layer
    end
      
  model.start_operation "Points Cloud Input"
  ss = model.selection
  ss.clear
  
  # Retrieve component definition if necessary
  if $point_type != "Construction point"
    comp_def =  Sketchup.active_model.definitions[$point_type]
  end
  
    # Process input file
     np = 1
    tableau_lignes.each do |line|
      current_point = line.split( pattern=@delim_options)
      if indx == nil
        x = (current_point[0].to_f.to_l * $mult)
        y = (current_point[1].to_f.to_l * $mult)
        z = (current_point[2].to_f.to_l * $mult)
        else
        x = (current_point[indx].to_f.to_l * $mult)
        y = (current_point[indy].to_f.to_l * $mult)
        z = (current_point[indz].to_f.to_l * $mult)
      end
      
      case flatten
        when "No"
          pt = Geom::Point3d.new(x,y,z)
        when "Yes"
          pt = Geom::Point3d.new(x,y,0.0)
      end
      Sketchup.set_status_text("Adding point " + np.to_s + "/ " + npts.to_s)
      if $point_type == "Construction point"
        cp = Sketchup.active_model.active_entities.add_cpoint pt
        ss.add cp
      else
        tr = Geom::Transformation.new pt
        Sketchup.active_model.active_entities.add_instance comp_def, tr
      end
      np = np + 1
    end
  
  # Zoom extents to points cloud
  view = model.active_view
  view.zoom_extents
  
  # Triangulation option
  answer = UI.messagebox("Would you like to triangulate the points cloud ?", MB_YESNOCANCEL, "Triangulation")
  case answer
    when 6
      triangulate_cloud
    when 7
      UI.messagebox((np-1).to_s + " points imported.")
    when 2
      UI.messagebox((np-1).to_s + " points imported.")
  end
  
  # Restores previous layer
  model.active_layer = start_layer
  # Close the file
  points_cloud_File.close
  # Command end and undo
  model.commit_operation 
  
  else
    UI.messagebox "No input file selected."
  end
end

def collect_components
model = Sketchup.active_model
components = model.definitions
list = []
components.each do |c|
  if not c.name["Group"] and not c.name["Image"]
    list.push(c.name)
  end
end
list
end

def collect_layers
list = []
model = Sketchup.active_model
layers = model.layers
layers.each do |l|
    list.push(l.name)
end
list
end

def triangulate_cloud
z = 0
triangles = []
points_table = []
points_table = points_to_array(z)
if points_table
  triangles = triangulate(points_table)
end
i = 0
0.upto( triangles.length - 1 ) do |something|
  indices = triangles[i]
  ind1 = indices[0]
  ind2 = indices[1]
  ind3 = indices[2]
  Sketchup.set_status_text("Adding triangle: "  + i.to_s + " / " + triangles.length.to_s )
  Sketchup.active_model.active_entities.add_face( points_table[ind3], points_table[ind2], points_table[ind1] )
  i = i + 1
end
end

def points_to_array(z)
  z = 1
  model = Sketchup.active_model
  model.start_operation "Points triangulation"
  entities = model.active_entities
  ss = model.selection
  if ss.empty? 
    UI.messagebox("No points to triangulate !")
    return nil
  end
  # Selection error checking: eliminates the non-cpoint objects
  look = 1
  i = 0
  0.upto( ss.length - 1) do |look|
    if( not ss[i].kind_of? Sketchup::ConstructionPoint )
      look = 0
    end
    i = i + 1
  end

  if(look == 0)
    UI.messagebox("Some object(s) of the selection aren't construction points.")
    return nil
  end
  
  i = 0
  # Fill array of 3D_points with selection
  points_array = []
    # read the points in the file and create the vertex array
    for ind in (0..(ss.length - 1))
      cpt = ss[i]
      x = cpt.position[0].to_f
      y = cpt.position[1].to_f
      z = cpt.position[2].to_f
      points_array[ind] = [x, y ,z]
      i = i + 1
    end
  points_array.uniq!
  return points_array
end

# Credit to Paul Bourke (pbourke@swin.edu.au) for original Fortran 77 Program :))
# August 2004 - Conversion and adaptation to Ruby by Carlos Fal (carlosfale@sapo.pt)
# September 2004 - Updated by Carlos Fal
#
# You can use this code, for non commercial purposes, however you like, providing the above credits remain in tact

# Return two logical values, first is TRUE if the point (xp, yp) lies inside the circumcircle made up by points (x1, y1), (x2, y2) and (x3, y3)
# and FALSE if not, second is TRUE if xc + r < xp and FALSE if not
# NOTE: A point on the edge is inside the circumcircle
def incircum(xp, yp, x1, y1, x2, y2, x3, y3)

  eps = 0.000001
  res = [FALSE, FALSE]

  if (y1 - y2).abs >= eps || (y2 - y3).abs >= eps
    if (y2 - y1).abs < eps
      m2 = -(x3 - x2) / (y3 - y2)
      mx2 = (x2 + x3) / 2
      my2 = (y2 + y3) / 2
      xc = (x1 + x2) / 2
      yc = m2 * (xc - mx2) + my2
    elsif (y3 - y2).abs < eps
      m1 = -(x2 - x1) / (y2 - y1)
      mx1 = (x1 + x2) / 2
      my1 = (y1 + y2) / 2
      xc = (x2 + x3) / 2
      yc = m1 * (xc - mx1) + my1
    else
      m1 = -(x2 - x1) / (y2 - y1)
      m2 = -(x3 - x2) / (y3 - y2)
      mx1 = (x1 + x2) / 2
      mx2 = (x2 + x3) / 2
      my1 = (y1 + y2) / 2
      my2 = (y2 + y3) / 2
      if (m1 - m2) == 0
        xc = (x1 + x2 + x3) / 3
        yc = (y1 + y2 + y3) / 3
      else
        xc = (m1 * mx1 - m2 * mx2 + my2 - my1) / (m1 - m2)
        yc = m1 * (xc - mx1) + my1
      end
    end

    dx = x2 - xc
    dy = y2 - yc
    rsqr = dx * dx + dy * dy
    r = Math.sqrt(rsqr)

    dx = xp - xc
    dy = yp - yc
    drsqr = dx * dx + dy * dy
    if drsqr < rsqr
      res[0] = TRUE
    end
    
    if xc + r < xp
      res[1] = TRUE
    end

  end

  return res

end

# Takes as input a array with NVERT lines, each line is a array with three values x, y, and z (vertex[j][0] = x value of j + 1 component) and
# return a array with NTRI lines, each line is a array with three values i, j, and k (each value is the index of a point in the input array (vertex array))
# i is the index of the first point, j is the index of the second point and k is the index of the third point,
def triangulate(vert)

  # Sort the input array in x values
  vert.sort!
  nvert = vert.length
  triang = Array.new
  edges = Array.new
  complete = Array.new

  # Verbose
  Sketchup.set_status_text("Starting triangulation of " + nvert.to_s + " Points")

  # Find the minimum and maximum vertex bounds. This is to allow calculation of the bounding triagle
  xmin = vert[0][0]
  ymin = vert[0][1]
  xmax = xmin
  ymax = ymin

  for i in (2..nvert)
    x1 = vert[i - 1][0]
    y1 = vert[i - 1][1]
    xmin = [x1, xmin].min
    xmax = [x1, xmax].max
    ymin = [y1, ymin].min
    ymax = [y1, ymax].max
  end

  dx = xmax - xmin
  dy = ymax - ymin
  dmax = [dx, dy].max
  xmid = (xmin + xmax) / 2
  ymid = (ymin + ymax) / 2

  # Set up the supertriangle. This is a triangle which encompasses all the sample points. The supertriangle coordinates are added to the end
  # of the vertex list. The supertriangle is the first triangle in the triangles list.
  p1 = nvert + 1
  p2 = nvert + 2
  p3 = nvert + 3
  vert[p1 - 1] = [xmid - 2 * dmax, ymid - dmax, 0]
  vert[p2 - 1] = [xmid, ymid + 2 * dmax, 0]
  vert[p3 - 1] = [xmid + 2 * dmax, ymid - dmax, 0]
  triang[0] = [p1 - 1, p2 - 1, p3 - 1]
  complete[0] = FALSE
  ntri = 1

  # Include each point one at a time into the exixsting mesh
  for i in (1..nvert)
    xp = vert[i - 1][0]
    yp = vert[i - 1][1]
    nedge = 0
    
  # Verbose  
  Sketchup.set_status_text("Triangulating point " + i.to_s + " / " + nvert.to_s )
  
    # Set up the edge buffer. If the point (xp, yp) lies inside the circumcircle then the three edges of that triangle are added to the edge buffer.
    j = 0
    while j < ntri
      j = j +1
      if complete[j - 1] != TRUE
        p1 = triang[j - 1][0]
	p2 = triang[j - 1][1]
	p3 = triang[j - 1][2]
	x1 = vert[p1][0]
	y1 = vert[p1][1]
	x2 = vert[p2][0]
	y2 = vert[p2][1]
	x3 = vert[p3][0]
	y3 = vert[p3][1]
	inc = incircum(xp, yp, x1, y1, x2, y2, x3, y3)
	if inc[1] == TRUE
	  complete[j - 1] = TRUE
	else
	  if inc[0] == TRUE
	    edges[nedge] = [p1, p2]
	    edges[nedge + 1] = [p2, p3]
	    edges[nedge + 2] = [p3, p1]
	    nedge = nedge + 3
	    triang[j - 1] = triang[ntri - 1]
	    complete[j - 1] = complete[ntri - 1]
	    j = j - 1
	    ntri = ntri - 1
	  end
	end
      end
    end

    # Tag multiple edges
    # NOTE: if all triangles are specified anticlockwise then all interior edges are pointing in direction.
    for j in (1..nedge - 1)
      if edges[j - 1][0] != -1 || edges[j - 1][1] != -1
        for k in ((j + 1)..nedge)
          if edges[k - 1][0] != -1 || edges[k - 1][1] != -1
	    if edges[j - 1][0] == edges[k - 1][1]
	      if edges[j - 1][1] == edges[k - 1][0]
	        edges[j - 1] = [-1, -1]
		edges[k - 1] = [-1, -1]
	      end
	    end
	  end
        end
      end
    end

    # Form new triangles for the current point. Skipping over any tagged adges. All edges are arranged in clockwise order.
    for j in (1..nedge)
      if edges[j - 1][0] != -1 || edges[j - 1][1] != -1
        ntri = ntri + 1
	triang[ntri - 1] = [edges[j - 1][0], edges[j - 1][1], i - 1]
	complete[ntri - 1] = FALSE
      end
    end
  end

  # Remove triangles with supertriangle vertices. These are triangles which have a vertex number greater than NVERT.
  i = 0
  while i < ntri
    i = i + 1
    if triang[i - 1][0] > nvert - 1 || triang[i - 1][1] > nvert - 1 || triang[i - 1][2] > nvert - 1
      triang[i - 1] = triang[ntri - 1]
      i = i - 1
      ntri = ntri - 1
    end
  end
  
  return triang[0..ntri - 1]

end
# ----------------------------------------------------------- Menu item
if( not file_loaded?("cloud.rb") )
  add_separator_to_menu("Plugins")
  UI.menu("Plugins").add_item("Import points cloud") { import_points_cloud }
end

# -----------------------------------------------------------------------------
file_loaded("cloud.rb")